// A SpreadSheet-type Control

// Copyright (C) 1999, 2000 by Makoto Nagata, MicroNova <mnagata@micronova.com>.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//
// Visit http://www.micronova.com for updates

import netscape.util.*;
import netscape.application.*;

import com.micronova.mnfc.*;

class SpreadSheetEditor extends TextField
{
    final KeyEvent lastEvent = new KeyEvent();

    public void keyDown(KeyEvent event)
    {
        lastEvent.key = event.key;
        lastEvent.modifiers = event.modifiers;

        switch (lastEvent.key)
        {
            case KeyEvent.ESCAPE_KEY:
            case KeyEvent.DOWN_ARROW_KEY:
            case KeyEvent.UP_ARROW_KEY:
            case KeyEvent.PAGE_DOWN_KEY:
            case KeyEvent.PAGE_UP_KEY:
            {
                event.key = KeyEvent.RETURN_KEY;
            }
            break;
        }

        super.keyDown(event);
    }
}

class SpreadSheetCell
{
    static Color[] colors = {Color.black, Color.red, Color.blue, Color.white, Color.yellow};
    static Font[] fonts = {Font.defaultFont(), Font.fontNamed("Helvetica:BOLD:16"), Font.fontNamed("TmsRmn:ITALIC:12"), Font.fontNamed("Courier:PLAIN:10")};

    static int[] justifications = {Graphics.CENTERED, Graphics.LEFT_JUSTIFIED, Graphics.RIGHT_JUSTIFIED};

    String value;
    int colorIndex;
    int fontIndex;
    int justificationIndex;
    Rect extent;
    View view;

    SpreadSheetCell()
    {
        value = "";
        colorIndex = 0;
        fontIndex = 0;
        justificationIndex = 0;
        extent = null;
        view = null;
    }
}

class SpreadSheet extends IGridView implements IGridDef, TextFieldOwner, Target
{
    int xCount;
    int yCount;

    int[] xPosition;
    int[] yPosition;

    Hashtable values = new Hashtable();

    IGridPane main;
    IGridPane top;
    IGridPane left;

    Point selectionBase = new Point();
    Point selectionPoint = new Point();
    Rect selection = new Rect();
    Rect newSelection = new Rect();

    int resizeIndex;
    int resizeOffset;

    Rect tempRect = new Rect();

    SpreadSheetEditor textField = new SpreadSheetEditor();

    SpreadSheetCell getCell(Point p)
    {
        SpreadSheetCell cell = (SpreadSheetCell)values.get(p);

        if (cell == null)
        {
            return defaultCell;
        }
        else
        {
            return cell;
        }
    }

    /** Draws text in a cell.  Must be called in drawCell(). */

    public final void drawText(IGridPane iGridPane, String s, Font font, Color color, int alignment, Border border)
    {
        if (font != null)
        {
            iGridPane.__g.setFont(font);
        }

        if (border != null)
        {
            border.drawInRect(iGridPane.__g, iGridPane.__cellRect);
        }

        if (color != null)
        {
            iGridPane.__g.setColor(color);
        }

        if (s != null)
        {
            iGridPane.__g.drawStringInRect(s, iGridPane.__cellRect, alignment);
        }
    }

    void updateSelection()
    {
        int lowX = selectionBase.x;
        int lowY = selectionBase.y;
                        
        int sWidth = selectionPoint.x - selectionBase.x;
        int sHeight = selectionPoint.y - selectionBase.y;
                        
        if (sWidth < 0)
        {
            lowX = selectionPoint.x;
            sWidth = -sWidth;
        }
                        
        if (sHeight < 0)
        {
            lowY = selectionPoint.y;
            sHeight = -sHeight;
        }
        
        newSelection.setBounds(lowX, lowY, sWidth + 1, sHeight + 1);                
        extendSubGrid(newSelection);

        if (!selection.equals(newSelection))
        {
            updateCells(selection);
            updateCells(newSelection);
        }

        selection.setBounds(newSelection);
    }

    void beginEdit(KeyEvent event)
    {
        SpreadSheetCell cell = getCell(selectionBase);

        textField.setStringValue(cell.value);

        textField.setJustification(SpreadSheetCell.justifications[cell.justificationIndex]);

        textField.setTextColor(SpreadSheetCell.colors[cell.colorIndex]);

        textField.setFont(SpreadSheetCell.fonts[cell.fontIndex]);

        textField.selectRange(new Range(0, cell.value.length()));

        Rect cellRect = new Rect();

        main.computeCellRect(selection, cellRect);

        Rect r = main.convertRectToView(this, cellRect);
        
        ContainerView mainContainer = containerViewArray[main.paneIndex];

        Rect mainContentRect = mainContainer.convertRectToView(this, mainContainer.interiorRect());

        r = r.intersectionRect(mainContentRect);

        textField.setBounds(r);
        
        textField.setFocusedView();

        if (event != null)
        {
            textField.keyDown(event);
        }
    }

    void endEdit()
    {
        KeyEvent lastEvent = ((SpreadSheetEditor)textField).lastEvent;

        if (lastEvent.key != KeyEvent.ESCAPE_KEY)
        {
            String newValue = textField.stringValue();

            for (int y = selection.y; y < selection.y + selection.height; y ++)
            {
                for (int x = selection.x; x < selection.x + selection.width; x ++)
                {
                    Point p = new Point(x, y);
                    SpreadSheetCell cell = getCell(p);

                    if (cell == defaultCell)
                    {
                        cell = new SpreadSheetCell();
                    }
                    
                    cell.value = newValue;
                    
                    values.put(p, cell);
                }
            }
        }

        textField.sizeTo(0, 0);

        rootView().setFocusedView(main);

        updateCells(selection);

        if (moveSelectionOnKey(lastEvent))
        {
            beginEdit(null);
        }
    }

    boolean moveSelectionOnKey(KeyEvent event)
    {
        int xIncrement = 0;
        int yIncrement = 0;

        boolean isShift = event.isShiftKeyDown();

        switch (event.key)
        {
            case KeyEvent.TAB_KEY:
            {
                if (isShift)
                {
                    xIncrement = -1;
                }
                else
                {
                    xIncrement = 1;
                }

                isShift = false;
            }
            break;

            case KeyEvent.DOWN_ARROW_KEY:
            {
                yIncrement = 1;
            }
            break;

            case KeyEvent.UP_ARROW_KEY:
            {
                yIncrement = -1;
            }
            break;
                            
            case KeyEvent.LEFT_ARROW_KEY:
            {
                xIncrement = -1;
            }
            break;

            case KeyEvent.RIGHT_ARROW_KEY:
            {
                xIncrement = 1;
            }
            break;

            case KeyEvent.PAGE_DOWN_KEY:
            {
                yIncrement = 10;
            }
            break;

            case KeyEvent.PAGE_UP_KEY:
            {
                yIncrement = -10;
            }
            break;

            case KeyEvent.HOME_KEY:
            {
                xIncrement = -xCount;
                yIncrement = -yCount;
            }
            break;
            
            case KeyEvent.END_KEY:
            {
                xIncrement = xCount;
                yIncrement = yCount;
            }
            break;
        }

        if ((xIncrement != 0) || (yIncrement != 0))
        {
            moveSelection(xIncrement, yIncrement, isShift);
            return true;
        }
        else
        {
            return false;
        }
    }
    
    void moveSelection(int xIncrement, int yIncrement, boolean isShift)
    {
        Point p = new Point(selectionPoint);
        Rect r = new Rect(selectionPoint.x, selectionPoint.y, 1, 1);

        main.iGridDef.extendCell(r);

        if (xIncrement > 0)
        {
            xIncrement += (r.width - 1);
        }

        if (yIncrement > 0)
        {
            yIncrement += (r.height - 1);
        }

        selectionPoint.x += xIncrement;
        selectionPoint.y += yIncrement;

        main.adjustCellIndex(selectionPoint);

        if (!isShift)
        {
            selectionBase.x = selectionPoint.x;
            selectionBase.y = selectionPoint.y;
        }

        scrollCellToVisible(selectionPoint);
        updateSelection();
    }

    SpreadSheet()
    {
        super();

        this.xCount = 50;
        this.yCount = 50;

        this.xPosition = new int[this.xCount + 1];
        this.yPosition = new int[this.yCount + 1];

        for (int i = 0; i <= this.xCount; i ++)
        {
            this.xPosition[i] = i * 80;
        }

        for (int i = 0; i <= this.yCount; i ++)
        {
            this.yPosition[i] = i * 18;
        }

        selection = new Rect(1, 1, 1, 1);
        selectionBase = new Point(1, 1);

        this.setIGridDef(this);

        main = this.iGridPaneArray[IGridView.MIDDLECENTER];
        top = this.iGridPaneArray[IGridView.TOPCENTER];
        left = this.iGridPaneArray[IGridView.MIDDLELEFT];
        
        main.setDelegateFlag(IView.MOUSE, true);
        main.setDelegateFlag(IView.KEY + IView.FOCUS, true);
        main.setDelegateFlag(IView.GET_TOOL_TIP_TEXT, true);
        main._mouseDown = true;
        main._wantsAutoscrollEvents = true;
        IToolTip.registerToolTip(main, null);

        top.setDelegateFlag(IView.MOUSE, true);
        top.setDelegateFlag(IView.CURSOR_FOR_POINT, true);

        left.setDelegateFlag(IView.MOUSE, true);
        left.setDelegateFlag(IView.CURSOR_FOR_POINT, true);

        textField.setOwner(this);
        textField.setJustification(Graphics.CENTERED);
        addSubview(textField);

        setBuffered(true);
    }

    public int getXCount()
    {
        return xCount;
    }

    public int getYCount()
    {
        return yCount;
    }
    
    public int getXPosition(int index)
    {
        return this.xPosition[index];
    }

    public int getYPosition(int index)
    {
        return this.yPosition[index];
    }

    public boolean extendCell(Rect subGrid)
    {
        Point p = new Point(subGrid.x, subGrid.y);

        SpreadSheetCell cell = getCell(p);

        if (cell != null)
        {
            if (cell.extent != null)
            {
                subGrid.setBounds(cell.extent);
                return true;
            }
        }

        return false;
    }

    SpreadSheetCell defaultCell = new SpreadSheetCell();

    public View drawCell(IGridPane iGridPane)
    {
        iGridPane.__g.setColor(Color.lightGray);

        iGridPane.__g.fillRect(iGridPane.__cellRect);

        if (iGridPane.paneIndex == IGridView.TOPCENTER)
        {
            drawText(iGridPane, "" + iGridPane.__index.x, null, SpreadSheetCell.colors[defaultCell.colorIndex], Graphics.CENTERED, BezelBorder.raisedBezel());

            return null;
        }
        else if (iGridPane.paneIndex == IGridView.MIDDLELEFT)
        {
            drawText(iGridPane, "" + iGridPane.__index.y, null, SpreadSheetCell.colors[defaultCell.colorIndex], Graphics.CENTERED, BezelBorder.raisedBezel());

            return null;
        }
            
        SpreadSheetCell cell = getCell(iGridPane.__index);

        Border border = LineBorder.grayLine();

        if (selection.contains(iGridPane.__index))
        {
            if (iGridPane.__index.equals(selectionBase))
            {
                border = BezelBorder.loweredBezel();
            }
            else
            {
                border = BezelBorder.groovedBezel();
            }
        }
        else if ((iGridPane.__index.x == 0) || (iGridPane.__index.y == 0))
        {
            border = BezelBorder.raisedBezel();
        }

        String s = cell.value;

        if (s == null)
        {
            s = "";
        }

        drawText(iGridPane, s, SpreadSheetCell.fonts[cell.fontIndex], SpreadSheetCell.colors[cell.colorIndex], SpreadSheetCell.justifications[cell.justificationIndex], border);

        return null;
    }

    public Rect getPaneSubGrid(int paneIndex)
    {
        return IGridView.computeStandardPaneSubGrid(this, paneIndex, 1, 0, 1, 0, new Rect());
    }

    void drawTempRect(IGridPane iGridPane)
    {
        Graphics g = iGridPane.iGridView.createGraphics();
        
        g.setXORMode(Color.lightGray);

        g.drawRect(tempRect);

        g.dispose();
    }

    public void delegatePane(int flag, IGridPane iGridPane)
    {
        switch (iGridPane.paneIndex)
        {
            case IGridView.MIDDLECENTER:
            {
                int rangeIndex = iGridPane.adjustCellIndex(iGridPane.__index);

                switch (flag)
                {
                    case IView.MOUSE_DRAGGED:
                    case IView.MOUSE_DOWN:
                    {
                        rootView().setFocusedView(iGridPane);

                        if (flag == IView.MOUSE_DOWN)
                        {
                            selectionBase.x = iGridPane.__index.x;
                            selectionBase.y = iGridPane.__index.y;
                            
                            iGridPane.adjustCellIndex(selectionBase);
                        }

                        selectionPoint.x = iGridPane.__index.x;
                        selectionPoint.y = iGridPane.__index.y;

                        iGridPane.adjustCellIndex(selectionPoint);

                        iGridPane.iGridView.scrollCellToVisible(selectionPoint);
                        updateSelection();

                        iGridPane.rootView().setFocusedView(iGridPane);
                    }
                    break;

                    case IView.GET_TOOL_TIP_TEXT:
                    {
                        iGridPane.computeCellRect(new Rect(iGridPane.__index.x, iGridPane.__index.y, 1, 1), iGridPane.__cellRect);

                        iGridPane.__cellRect.growBy(-2, -2);

                        if (iGridPane.__cellRect.contains(iGridPane.__x, iGridPane.__y))
                        {
                            SpreadSheetCell cell = getCell(iGridPane.__index);

                            if (!cell.value.equals(""))
                            {
                                iGridPane._getToolTipText = cell.value;
                            }
                            else
                            {
                                iGridPane._getToolTipText = null;
                            }
                        }
                        else
                        {
                            iGridPane._getToolTipText = null;
                        }
                    }
                    break;

                    case IView.KEY_DOWN:
                    {
                        KeyEvent keyEvent = (KeyEvent)iGridPane.__event;

                        if (!moveSelectionOnKey(keyEvent))
                        {
                            beginEdit(keyEvent);
                        }
                    }
                    break;
                }
            }
            break;

            case IGridView.TOPCENTER:
            {                        
                iGridPane.adjustCellIndex(iGridPane.__index);

                switch (flag)
                {
                    case IView.CURSOR_FOR_POINT:
                    case IView.MOUSE_DOWN:
                    {
                        int endPosition = iGridPane.getXPosition(iGridPane.__index.x + 1);
                        int endDistance = endPosition - iGridPane.__x;

                        if ((endDistance > 0) && (endDistance < 5))
                        {
                            iGridPane._cursorForPoint = View.E_RESIZE_CURSOR;
                            iGridPane._mouseDown = true;

                            if (flag == IView.MOUSE_DOWN)
                            {
                                resizeOffset = endDistance;
                                resizeIndex = iGridPane.__index.x;

                                Rect r = new Rect(iGridPane.__index.x, iGridPane.__index.y, 1, 1);
                                iGridPane.computeCellRect(r, tempRect);
                            
                                iGridPane.convertRectToView(iGridPane.iGridView, tempRect, tempRect);

                                tempRect.y = iGridPane.iGridView.containerViewArray[TOPCENTER].y();
                                tempRect.height = iGridPane.iGridView.containerViewArray[TOPCENTER].height() + iGridPane.iGridView.containerViewArray[MIDDLECENTER].height();
                                drawTempRect(iGridPane);
                            }
                        }
                        else
                        {
                            iGridPane._cursorForPoint = View.ARROW_CURSOR;
                            iGridPane._mouseDown = false;
                        }

                    }
                    break;

                    case IView.MOUSE_DRAGGED:
                    {
                        drawTempRect(iGridPane);

                        MouseEvent eventParent = (MouseEvent)iGridPane.convertEventToView(iGridPane.iGridView, (MouseEvent)iGridPane.__event);
                        
                        if (eventParent.x <= tempRect.x + 10)
                        {
                            eventParent.x = tempRect.x + 10;
                        }

                        tempRect.width = eventParent.x - tempRect.x + resizeOffset;
                        drawTempRect(iGridPane);
                    }
                    break;

                    case IView.MOUSE_UP:
                    {
                        drawTempRect(iGridPane);

                        int width = tempRect.width;
                        
                        int delta = width - (getXPosition(resizeIndex + 1) - getXPosition(resizeIndex));

                        for (int c = resizeIndex + 1; c <= xCount; c ++)
                        {
                            xPosition[c] += delta;
                        }

                        Rect resizeCells = new Rect(resizeIndex - 1, 0, xCount - resizeIndex + 1, yCount);

                        extendSubGrid(resizeCells);
                   
                        updateCells(resizeCells);
                    }
                    break;
                }
            }
            break;

            case IGridView.MIDDLELEFT:
            {                        
                iGridPane.adjustCellIndex(iGridPane.__index);

                switch (flag)
                {
                    case IView.CURSOR_FOR_POINT:
                    case IView.MOUSE_DOWN:
                    {
                        int endPosition = iGridPane.getYPosition(iGridPane.__index.y + 1);
                        int endDistance = endPosition - iGridPane.__y;

                        if ((endDistance > 0) && (endDistance < 5))
                        {
                            iGridPane._cursorForPoint = View.S_RESIZE_CURSOR;
                            iGridPane._mouseDown = true;

                            if (flag == IView.MOUSE_DOWN)
                            {
                                resizeOffset = endDistance;
                                resizeIndex = iGridPane.__index.y;

                                Rect r = new Rect(iGridPane.__index.x, iGridPane.__index.y, 1, 1);
                                iGridPane.computeCellRect(r, tempRect);
                            
                                iGridPane.convertRectToView(iGridPane.iGridView, tempRect, tempRect);

                                tempRect.x = iGridPane.iGridView.containerViewArray[MIDDLELEFT].x();
                                tempRect.width = iGridPane.iGridView.containerViewArray[MIDDLELEFT].width() + iGridPane.iGridView.containerViewArray[MIDDLECENTER].width();
                                drawTempRect(iGridPane);
                            }
                        }
                        else
                        {
                            iGridPane._cursorForPoint = View.ARROW_CURSOR;
                            iGridPane._mouseDown = false;
                        }

                    }
                    break;

                    case IView.MOUSE_DRAGGED:
                    {
                        drawTempRect(iGridPane);

                        MouseEvent eventParent = (MouseEvent)iGridPane.convertEventToView(iGridPane.iGridView, (MouseEvent)iGridPane.__event);
                        
                        if (eventParent.y <= tempRect.y + 10)
                        {
                            eventParent.y = tempRect.y + 10;
                        }

                        tempRect.height = eventParent.y - tempRect.y + resizeOffset;
                        drawTempRect(iGridPane);
                    }
                    break;

                    case IView.MOUSE_UP:
                    {
                        drawTempRect(iGridPane);

                        int height = tempRect.height;
                        
                        int delta = height - (getYPosition(resizeIndex + 1) - getYPosition(resizeIndex));

                        for (int c = resizeIndex + 1; c <= yCount; c ++)
                        {
                            yPosition[c] += delta;
                        }

                        updateCells(new Rect(0, resizeIndex - 1, xCount, yCount - resizeIndex + 1));
                    }
                    break;
                }
            }
            break;

        }
    }

    public void textEditingDidBegin(TextField textField)
    {
    }

    public void textWasModified(TextField textField)
    {
    }

    public boolean textEditingWillEnd(TextField textField, int endCondition, boolean contentsChanged)
    {
        return true;
    }

    public void textEditingDidEnd(TextField textField, int endCondition, boolean contentsChanged)
    {
        endEdit();
    }

    public void performCommand(String command, Object object)
    {
        for (int y = selection.y; y < selection.y + selection.height; y ++)
        {
            for (int x = selection.x; x < selection.x + selection.width; x ++)
            {
                Point p = new Point(x, y);
                SpreadSheetCell cell = getCell(p);

                if (cell == defaultCell)
                {
                    cell = new SpreadSheetCell();
                }

                if (command.equals("COLOR"))
                {
                    cell.colorIndex = (cell.colorIndex + 1) % SpreadSheetCell.colors.length;
                }
                else if (command.equals("FONT"))
                {
                    cell.fontIndex = (cell.fontIndex + 1) % SpreadSheetCell.fonts.length;
                }
                else if (command.equals("JUSTIFY"))
                {
                    cell.justificationIndex = (cell.justificationIndex + 1) % SpreadSheetCell.justifications.length;
                }
                else if (command.equals("GROUP"))
                {
                    cell.extent = new Rect(selection);
                }
                else if (command.equals("UNGROUP"))
                {
                    cell.extent = null;
                }

                values.put(p, cell);
            }
        }

        if (command.equals("GROUP") || command.equals("UNGROUP"))
        {
            selectionBase.x = selection.x;
            selectionBase.y = selection.y;

            selectionPoint.x = selection.x;
            selectionPoint.y = selection.y;

            updateSelection();
        }

        updateCells(selection);

        super.performCommand(command, object);
    }
}

public class Test extends IApplication
{
    public void init()
    {
        SpreadSheet spreadSheet = new SpreadSheet();

        InternalWindow window = new InternalWindow(0, 0, 400, 200);

        window.setResizable(true);

        Menu menu = new Menu();

        MenuItem editList = menu.addItemWithSubmenu("Edit");
        
        editList.submenu().addItem("Change Color", "COLOR", spreadSheet);
        editList.submenu().addItem("Change Font", "FONT", spreadSheet);
        editList.submenu().addItem("Change Justification", "JUSTIFY", spreadSheet);
        editList.submenu().addItem("Group Cells", "GROUP", spreadSheet);
        editList.submenu().addItem("Ungroup Cells", "UNGROUP", spreadSheet);

        window.setMenuView(new MenuView(menu));

        Size size = window.contentSize();

        spreadSheet.sizeTo(size.width, size.height);

        window.addSubview(spreadSheet);

        spreadSheet.setHorizResizeInstruction(View.WIDTH_CAN_CHANGE);
        spreadSheet.setVertResizeInstruction(View.HEIGHT_CAN_CHANGE);

        window.center();
        
        window.show();
    }
}
