// Grid Definition Interface

// Copyright (C) 1999, 2000 by Makoto Nagata, MicroNova <mnagata@micronova.com>.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//
// Visit http://www.micronova.com for updates

package com.micronova.mnfc;

import netscape.util.*;
import netscape.application.*;

/**

   <P>IGridDef defines a 'grid' (rows and columns of 'cell's).  A grid is defined
   by specifying the following:

   <P>
   <UI>
   <LI>M: Number of 'cell's in X-direction</LI>
   <LI>N: Number of 'cell's in Y-direction</LI>
   <LI>x[0] < x[1] < ... < x[M]: (M + 1) boundary points in X-direction</LI>
   <LI>y[0] < y[1] < ... < y[N]: (N + 1) boundary points in Y-direction</LI>
   </UI>

   <P>Each cell is indicated by its 'index' (i, j) (i = 0, 1, ..., M - 1; j = 0, 1, ..., N - 1).  The cell at (i, j) is bounded by the rectangle (x, y, width, height) = (x[i], y[j], x[i + 1] - x[i], y[j + 1] - y[j]).  This pixel rectangle is called the 'cell rectangle'.

   <P>A 'subgrid' is a rectangular subset of cells; e.g., a subgrid (0, 0, 2, 3) (based at (0, 0), width 2, height 3) indicates the set of 6 cells at (0, 0), (1, 0), (0, 1), (1, 1), (0, 2), (1, 2).

   <P>A subgrid can be made into a single cell by extendCell().

   <P>For each cell, IGridDef specifies how to draw it and/or put a subview 
   (called a 'cell view') there in drawCell().

   An instance of <A HREF="com.micronova.mnfc.IGridPane.html">IGridPane</A> 
   actually draws a subgrid (called a 'pane') of IGridDef.  Each IGridPane has
   an index to indicate which subgrid it draws, and calls IGridDef's 
   getPaneRect() to find out which subgrid to draw.
   
   <P>IGridDef can also handle various events (MouseDown, etc.) happening
   in each IGridPane displaying it.

   * @author Makoto Nagata, MicroNova
   * @version 1.0

*/

public interface IGridDef
{
    /** Number of cells in X direction. */

    public int getXCount();

    /** Number of cells in Y direction. */

    public int getYCount();

    /** Position of xIndex-th boundary point (0 <= xIndex <= getXCount()). */

    public int getXPosition(int xIndex);

    /** Position of yIndex-th boundary point (0 <= yIndex <= getYCount()). */

    public int getYPosition(int yIndex);

    /** Extends a cell to have multiple row/column spans.  This method is called for each cell (i, j) with subGrid = (i, j, 1, 1).  In order to make a subgrid (x, y, w, h) into a single cell, set the subGrid to (w, y, w, h) for each cell (i, j) in the subgrid and return true.  For non-spanned cells, simply return false without modifying the subGrid. */

    public boolean extendCell(Rect subGrid);

    /** Draws a cell.  If this returns a non-null View, it is added to the IGridPane as a subview.  The following variables of iGridPane are set before this method is called:<BR>

     <UL>

     <LI>__g: Graphics object for drawing.
     <LI>__index: (i, j) if the cell to be drawn is at index (i, j)
     <LI>__subGrid: subgrid for index (i, j).  This is usually (i, j, 1, 1), but
     can be different if the cell is extended. (see extendCell()).
     <LI>__cellRect: pixel rectangle for the cell (in IGridPane's coordinates)
     <LI>__cellView: current cell view for this index.  null if no cellView has
     been created.
     </UL>

     <P>To have a cell view, check __cellView.  If __cellView is not null,
     simply update the view, size it according to __cellRect,  and return it. 
     If __cellView is null, create a new View, size it according to __cellRect,
     and return it.  If you don't want a cell view at the cell 
     (regardless of current __cellView), return null.

    */

    public View drawCell(IGridPane iGridPane);

    /** Defines the pane subgrid for each paneIndex.  IGridView defines 9 paneIndex values (IGridView.TOPLEFT ... IGridView.BOTTOMRIGHT). */

    public Rect getPaneSubGrid(int paneIndex);

    /** Method delegation from each IGridPane displaying this IGridDef (see <A HREF="com.micronova.mnfc.IView.html">IView</A>).  For MOUSE_XXX delegation, __index is set to the cell index under mouse. */

    public void delegatePane(int command, IGridPane iGridPane);
}
