// Grid Pane

// Copyright (C) 1999, 2000 by Makoto Nagata, MicroNova <mnagata@micronova.com>.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//
// Visit http://www.micronova.com for updates

package com.micronova.mnfc;

import netscape.util.*;
import netscape.application.*;

/** 

    An IGridPane displays one grid pane defined by <A HREF="com.micronova.mnfc.IGridDef.html">IGridDef</A>.
    <P>An IGridPane has a pane index (<A HREF="com.micronova.mnfc.IGridView.html">IGridView</A> defines 9 pane
    indices: TOPLEFT, TOPCENTER, TOPRIGHT, MIDDLELEFT,
    MIDDLECENTER, MIDDLERIGHT, BOTTOMLEFT, BOTTOMCENTER, 
    BOTTOMRIGHT) and displays the pane defined by IGridDef's
    getPaneSubGrid(paneIndex).
    
    <P>IGridPane inherits from <A HREF="com.micronova.mnfc.IView.html">IView</A>.
    It self-delegates DRAW_VIEW and calls IGridDef's drawCell() for each cell
    after setting up the delegate variables __subGrid, __index, __cellRect, and
    __cellView.  It re-delegates others to IGridDef's delegatePane().
    For MOUSE_XXX, it sets up __index to the cell index under mouse
    (possibly out of range) before calling delegatePane().

    * @author Makoto Nagata, MicroNova
    * @version 1.01

*/

public class IGridPane extends IView
{
    /** IGridDef this IGridPane displays. */

    public IGridDef iGridDef;

    /** Pane Index (IGridView.MIDDLELEFT, etc.) */

    public int paneIndex;

    /* IGridView which this IGridPane belongs to. */

    public IGridView iGridView;

    /** Must be set to true if this IGridPane has cellviews. */

    public boolean hasCellViews = false;

    /** Hashtable that maps index points to cellviews. */

    public Hashtable cellViewTable = new Hashtable();

    /** Hashtable that maps cellviews to index points.  This is not allocated
until getCellViewIndex() is called. */

    public Hashtable indexTable = null;

    /** Temporary variables. */

    private final Rect tempRect = new Rect();
    private final Rect tempSubGrid = new Rect();
    private final Point tempIndex = new Point();

    /** Delegate variable: current subgrid. */

    public final Rect __subGrid = new Rect();

    /** Delegate variable: current index. */

    public final Point __index = new Point();

    /** Delegate variable: current cell (pixel) rectangle. */

    public final Rect __cellRect = new Rect();

    /** Delegate variable: current cell view. */

    public View __cellView = null;

    /** Last cell index under mouse. */

    public final Point cellUnderMouse = new Point(-1, -1);

    /** Cached pane subgrid.   Set in setup().*/

    protected Rect paneSubGrid;

    /** Cached 'base point' in X direction (this pane's relative
        position in the whole grid defined by IGridDef).  Set in setup(). */

    protected int xBase;

    /** Cached 'base point' in Y direction (this pane's relative
        position in the whole grid defined by IGridDef).  Set in setup(). */

    protected int yBase;

    /** Returns the cell view at index p.  Returns null if no cell view is defined. */

    public View getCellView(Point p)
    {
        return (View)cellViewTable.get(p);
    }

    /** Returns the index for a given cell view. */

    public Point getCellViewIndex(View cellView)
    {
        if (indexTable == null)
        {
            indexTable = new Hashtable();
            
            Enumeration enum = cellViewTable.keys();

            while(enum.hasMoreElements())
            {
                Point p = (Point)enum.nextElement();
                View v = (View)cellViewTable.get(p);

                indexTable.put(v, p);
            }
        }

        return (Point)indexTable.get(cellView);
    }

    /** Sets cell view at index p.  'cellView' can be null. */

    public void setCellView(Point pIndex, View cellView)
    {
        if (cellView != null)
        {
            cellViewTable.put(new Point(pIndex), cellView);
            
            if (indexTable != null)
            {
                indexTable.put(cellView, pIndex);
            }
        }
        else
        {
            cellViewTable.remove(pIndex);

            if (indexTable != null)
            {
                indexTable.remove(cellView);
            }
        }
    }
    
    /** Constructs an IGridPane with given IGridDef and paneIndex. */

    public IGridPane(IGridDef iGridDef, int paneIndex)
    {
        super();

        this.iGridDef = iGridDef;
        this.paneIndex = paneIndex;
        
        setDelegateFlag(IView.DRAW_VIEW);
        
        setup();
    }

    /** Sets IGridDef.*/

    public void setIGridDef(IGridDef iGridDef)
    {
        this.iGridDef = iGridDef;

        setup();
    }

    /** Sets up cache variables (and resizes this IGridPane if necessary).  
        An IGridPane's size is always equal to that of the pane it is 
        displaying.  This is automatically called in drawView() and
        updateCells() so it is  sually not necessary to call this yourself. */

    public void setup()
    {
        paneSubGrid = iGridDef.getPaneSubGrid(paneIndex);

        xBase = iGridDef.getXPosition(paneSubGrid.x);
        yBase = iGridDef.getYPosition(paneSubGrid.y);

        int width = iGridDef.getXPosition(paneSubGrid.maxX()) - xBase;
        int height = iGridDef.getYPosition(paneSubGrid.maxY()) - yBase;

        if ((width() != width) || (height() != height))
        {
            sizeTo(width, height);
        }
    }

    /** Returns the X index for a pixel position x0, assuming it is in the range [lowIndex, highIndex]. */

    public int getXIndex(int x0, int lowIndex, int highIndex)
    {
        int x = x0 + xBase;

        if (x < iGridDef.getXPosition(lowIndex))
        {
            return lowIndex - 1;
        }
        else if (iGridDef.getXPosition(highIndex) <= x)
        {
            return highIndex;
        }
        else
        {
            while (lowIndex + 1L < highIndex)
            {
                int middleIndex = (lowIndex + highIndex) / 2;
     
                if (iGridDef.getXPosition(middleIndex) <= x)
                {
                    lowIndex = middleIndex;
                }
                else
                {
                    highIndex = middleIndex;
                }
            }
     
            return highIndex - 1;
        }
    }

    /** Returns the X index for a position x (requires setup()). */

    public int getXIndex(int x)
    {
        return getXIndex(x, paneSubGrid.x, paneSubGrid.x + paneSubGrid.width);
    }

    /** Returns the Y index for a position y0, assuming it is in the given range [lowIndex, highIndex] */

    public int getYIndex(int y0, int lowIndex, int highIndex)
    {
        int y = y0 + yBase;

        if (y < iGridDef.getYPosition(lowIndex))
        {
            return lowIndex - 1;
        }
        else if (iGridDef.getYPosition(highIndex) <= y)
        {
            return highIndex;
        }
        else
        {
            while (lowIndex + 1L < highIndex)
            {
                int middleIndex = (lowIndex + highIndex) / 2;
     
                if (iGridDef.getYPosition(middleIndex) <= y)
                {
                    lowIndex = middleIndex;
                }
                else
                {
                    highIndex = middleIndex;
                }
            }
     
            return highIndex - 1;
        }
    }

    /** Returns the Y index for a position y (requires setup()). */

    public int getYIndex(int y)
    {
        return getYIndex(y, paneSubGrid.y, paneSubGrid.y + paneSubGrid.height);
    }
    
    /** Returns the x position relative to this IGridPane for an index. */

    public int getXPosition(int xIndex)
    {
        return iGridDef.getXPosition(xIndex) - xBase;
    }

    /** Returns the y position relative to this IGridPane for an index. */

    public int getYPosition(int yIndex)
    {
        return iGridDef.getYPosition(yIndex) - yBase;
    }

    /** 
        Adjusts a (possibly out-of-range) index to a valid in-range index. 
        Also returns an integer flag to indicate whether the original
        index was in-range or not as follows:<BR>

        <UL>
        <LI>IGridView.TOPLEFT: x too small, y too small
        <LI>IGridView.TOPCENTER: x in range, y too small
        <LI>IGridView.TOPRIGHT: x too large, y too small
        <LI>IGridView.MIDDLELEFT: x too small, y in range
        <LI>IGridView.MIDDLECENTER: x in range, y in range
        <LI>IGridView.MIDDLERIGHT: x too large, y in range
        <LI>IGridView.BOTTOMLEFT: x too small, y too large
        <LI>IGridView.BOTTOMCENTER: x in range, y too large
        <LI>IGridView.BOTTOMRIGHT: x too large, y too large
        </UL>

    */

    public int adjustCellIndex(Point indexPoint)
    {
        int xRange = 1;
        int yRange = 1;

        if (indexPoint.x < paneSubGrid.x)
        {
            indexPoint.x = paneSubGrid.x;
            xRange = 0;
        }
        else if (indexPoint.x >= paneSubGrid.x + paneSubGrid.width)
        {
            indexPoint.x = paneSubGrid.x + paneSubGrid.width - 1;
            xRange = 2;
        }

        if (indexPoint.y < paneSubGrid.y)
        {
            indexPoint.y = paneSubGrid.y;
            yRange = 0;
        }
        else if (indexPoint.y >= paneSubGrid.y + paneSubGrid.height)
        {
            indexPoint.y = paneSubGrid.y + paneSubGrid.height - 1;
            yRange = 2;
        }

        tempSubGrid.setBounds(indexPoint.x, indexPoint.y, 1, 1);

        if (iGridDef.extendCell(tempSubGrid))
        {
            indexPoint.x = tempSubGrid.x;
            indexPoint.y = tempSubGrid.y;
        }

        return yRange * 3 + xRange;
    }

    /** Computes cell rectangle for given subGrid. */

    public void computeCellRect(Rect subGrid, Rect cellRect)
    {
        int x = iGridDef.getXPosition(subGrid.x);
        int y = iGridDef.getYPosition(subGrid.y);
        int w = iGridDef.getXPosition(subGrid.x + subGrid.width) - x;
        int h = iGridDef.getYPosition(subGrid.y + subGrid.height) - y;

        cellRect.setBounds(x - xBase, y - yBase, w, h);
    }

    /** Computes the smallest subgrid that contains given pixel rectangle. */

    public void computeClosureSubGrid(Rect rect, Rect subGrid)
    {
        int xLow = getXIndex(rect.x);
        int xHigh = getXIndex(rect.x + rect.width);
        int yLow = getYIndex(rect.y);
        int yHigh = getYIndex(rect.y + rect.height);

        subGrid.setBounds(xLow, yLow, xHigh - xLow + 1, yHigh - yLow + 1);
    }

    /** Computes the largest subgrid that is contained in given pixel rectangle. */

    public void computeInteriorSubGrid(Rect rect, Rect subGrid)
    {
        int xLow = getXIndex(rect.x);
        int xHigh = getXIndex(rect.x + rect.width) - 1;
        int yLow = getYIndex(rect.y);
        int yHigh = getYIndex(rect.y + rect.height) - 1;

        if (rect.x > getXPosition(xLow))
        {
            xLow ++;
        }

        if (rect.y >= getYPosition(yLow))
        {
            yLow ++;
        }

        subGrid.setBounds(xLow, yLow, xHigh - xLow + 1, yHigh - yLow + 1);
    }

    /** Updates all the cells in the given subGrid after processing current event. */

    public void updateCells(Rect subGrid)
    {
        setup();
        
        computeCellRect(subGrid, tempRect);

        addDirtyRect(tempRect);
    }

    /** Updates all the cells in the given subGrid immediately. */

    public void updateCellsNow(Rect subGrid)
    {
        setup();
        
        computeCellRect(subGrid, tempRect);

        draw(tempRect);
    }

    /** Draws a cell. */

    public void drawCell()
    {
        if (hasCellViews)
        {
            this.__cellView = getCellView(this.__index);

            if (this.__cellView != null)
            {
                this.__cellView.disableDrawing();
            }

            this.__cellView = iGridDef.drawCell(this);

            setCellView(this.__index, this.__cellView);
        }
        else
        {
            iGridDef.drawCell(this);
        }
    }

    /** Delegated drawView(). */

    private void _drawView()
    {
        Rect rect = __g.clipRect();

        if (rect == null)
        {
            tempRect.setBounds(0, 0, width(), height());
            rect = tempRect;
        }

        __g.setFont(Font.defaultFont());

        __g.pushState();

        setup();

        int yMinIndex = getYIndex(rect.y);
        int yMaxIndex = getYIndex(rect.y + rect.height - 1) + 1;
        int xMinIndex = getXIndex(rect.x);
        int xMaxIndex = getXIndex(rect.x + rect.width - 1) + 1;

        if (yMaxIndex > paneSubGrid.y + paneSubGrid.height)
        {
            yMaxIndex = paneSubGrid.y + paneSubGrid.height;
        }
        else if (yMaxIndex < paneSubGrid.y)
        {
            yMaxIndex = paneSubGrid.y;
        }

        if (xMaxIndex > paneSubGrid.x + paneSubGrid.width)
        {
            xMaxIndex = paneSubGrid.x + paneSubGrid.width;
        }
        else if (xMaxIndex < paneSubGrid.x)
        {
            xMaxIndex = paneSubGrid.x;
        }

        __subGrid.width = 1;
        __subGrid.height = 1;

        __index.y = yMinIndex;
        __subGrid.y = yMinIndex;

        int yLow = iGridDef.getYPosition(__index.y);

        for (int yHighIndex = yMinIndex + 1; yHighIndex <= yMaxIndex; yHighIndex ++)
        {
            int yHigh = iGridDef.getYPosition(yHighIndex);

            __index.x = xMinIndex;
            __subGrid.x = xMinIndex;

            int xLow = iGridDef.getXPosition(__index.x);

            for (int xHighIndex = xMinIndex + 1; xHighIndex <= xMaxIndex; xHighIndex ++)
            {
                int xHigh = iGridDef.getXPosition(xHighIndex);

                if (iGridDef.extendCell(__subGrid))
                {   
                    int xBaseIndex = __subGrid.x;

                    if (xBaseIndex < xMinIndex)
                    {
                        xBaseIndex = xMinIndex;
                    }

                    int yBaseIndex = __subGrid.y;

                    if (yBaseIndex < yMinIndex)
                    {
                        yBaseIndex = yMinIndex;
                    }

                    if ((__index.x == xBaseIndex) && (__index.y == yBaseIndex))
                    {
                        __cellRect.x = iGridDef.getXPosition(__subGrid.x);
                        __cellRect.y = iGridDef.getYPosition(__subGrid.y);
                        __cellRect.width = iGridDef.getXPosition(__subGrid.x + __subGrid.width) - __cellRect.x;
                        __cellRect.height = iGridDef.getYPosition(__subGrid.y + __subGrid.height) - __cellRect.y;

                        __cellRect.x -= xBase;
                        __cellRect.y -= yBase;

                        int xSave = __index.x;
                        int ySave = __index.y;
                        
                        __index.x = __subGrid.x;
                        __index.y = __subGrid.y;

                        drawCell();

                        __index.x = xSave;
                        __index.y = ySave;
                    }

                    __subGrid.x = __index.x;
                    __subGrid.y = __index.y;
                    __subGrid.width = 1;
                    __subGrid.height = 1;
                }
                else
                {
                    __cellRect.x = xLow - xBase;
                    __cellRect.y = yLow - yBase;
                    __cellRect.width = xHigh - xLow;
                    __cellRect.height = yHigh - yLow;

                    drawCell();
                }

                __index.x = xHighIndex;
                __subGrid.x = xHighIndex;
                xLow = xHigh;
            }

            __index.y = yHighIndex;
            __subGrid.y = yHighIndex;
            yLow = yHigh;
        }

        __g.popState();

        if (hasCellViews && (!cellViewTable.isEmpty()))
        {
            computeVisibleRect(tempRect);

            computeClosureSubGrid(tempRect, tempSubGrid);

            Vector s = subviews();

            while (!s.isEmpty())
            {
                View view = (View)s.removeLastElement();
            }
        
            for (tempIndex.y = tempSubGrid.y; tempIndex.y <= tempSubGrid.y + tempSubGrid.height; tempIndex.y ++)
            {
                for (tempIndex.x = tempSubGrid.x; tempIndex.x <= tempSubGrid.x + tempSubGrid.width; tempIndex.x ++)
                {
                    View view = getCellView(tempIndex);
                    
                    if (view != null)
                    {
                        addSubview(view);
                        view.reenableDrawing();
                    }
                }
            }
        }
    }

    /** IView delegation. */

    public void delegate(int flag, IView iView)
    {
        switch (flag)
        {
            case IView.DRAW_VIEW:
            {
                _drawView();
            }
            break;

            case IView.MOUSE_DOWN:
            case IView.MOUSE_UP:
            case IView.MOUSE_DRAGGED:
            case IView.MOUSE_MOVED:
            case IView.MOUSE_ENTERED:
            case IView.MOUSE_EXITED:
            case IView.GET_TOOL_TIP_TEXT:
            {
                __x = ((MouseEvent)__event).x;
                __y = ((MouseEvent)__event).y;

                if (flag == IView.MOUSE_EXITED)
                {
                    __index.x = cellUnderMouse.x;
                    __index.y = cellUnderMouse.y;

                    iGridDef.delegatePane(IView.MOUSE_EXITED, this);

                    cellUnderMouse.x = -1;
                    cellUnderMouse.y = -1;
                }
                else
                {
                    __index.x = getXIndex(__x);
                    __index.y = getYIndex(__y);
                
                    if ((flag == IView.MOUSE_MOVED) &&(!__index.equals(cellUnderMouse)))
                    {
                        int newX = __index.x;
                        int newY = __index.y;

                        __index.x = cellUnderMouse.x;
                        __index.y = cellUnderMouse.y;

                        iGridDef.delegatePane(IView.MOUSE_EXITED, this);

                        __index.x = newX;
                        __index.y = newY;

                        iGridDef.delegatePane(IView.MOUSE_ENTERED, this);
                    }

                    iGridDef.delegatePane(flag, this);

                    cellUnderMouse.x = __index.x;
                    cellUnderMouse.y = __index.y;
                }
            }
            break;

            default:
            {
                iGridDef.delegatePane(flag, this);
            }
            break;
        }
    }
}








