// Grid View

// Copyright (C) 1999, 2000 by Makoto Nagata, MicroNova <mnagata@micronova.com>.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//
// Visit http://www.micronova.com for updates

package com.micronova.mnfc;

import netscape.application.*;
import netscape.util.*;

/**

   IGridView is a scrollable grid control with 9 panes.  
   The grid displayed is defined by the IGridDef object.

   <P>IGridView consists of 9 panes; TOPLEFT (fixed size) 
   TOPCENTER (horizontally scrollable, fixed height), 
   TOPRIGHT (fixed size), 
   MIDDLELEFT (vertically scrollable, fixed width), 
   MIDDLECENTER (vertically and horizontally scrollable), 
   MIDDLERIGHT (vertically scrollable, fixed width), 
   BOTOTMLEFT (fixed size), 
   BOTTOMCENTER (horizontally scrollable, fixed height), 
   BOTTOMRIGHT (fixed size).

   A common Border separates the panes (default is BezelBorder.groovedBezel()).

   * @author Makoto Nagata, MicroNova
   * @version 1.0

*/

public class IGridView extends View implements LayoutManager, Target
{
    /** Pane index for the top-left pane. */

    public static final int TOPLEFT = 0;

    /** Pane index for the top-center pane. */

    public static final int TOPCENTER = 1;

    /** Pane index for the top-right pane. */

    public static final int TOPRIGHT = 2;

    /** Pane index for the middle-left pane. */

    public static final int MIDDLELEFT = 3;

    /** Pane index for the middle-center pane. */

    public static final int MIDDLECENTER = 4;

    /** Pane index for the middle-right pane. */

    public static final int MIDDLERIGHT = 5;

    /** Pane index for the bottom-left pane. */

    public static final int BOTTOMLEFT = 6;

    /** Pane index for the bottom-center pane. */

    public static final int BOTTOMCENTER = 7;

    /** Pane index for the bottom-right pane. */

    public static final int BOTTOMRIGHT = 8;

    /** Grid definition. */

    IGridDef iGridDef;

    /** Holds IGridPanes for each pane.  Each IGridPane is a subview of ScrollView which in turn in a subview of ContainerView. */

    public final IGridPane[] iGridPaneArray = new IGridPane[9];

    /** Holds ScrollViews for each pane. */

    public final ScrollView[] scrollViewArray = new ScrollView[9];

    /** Holds ContainerViews for each pane. */

    public final ContainerView[] containerViewArray = new ContainerView[9];

    /** Horizontal scroll bar. */

    public ScrollBar horizScrollBar;

    /** Vertical scroll bar. */

    public ScrollBar vertScrollBar;

    /** Pane border.   Default is BezelBorder.groovedBezel(). */

    public Border paneBorder = BezelBorder.groovedBezel();

    /** border.widthMargin(). */

    public int paneBorderWidth = 0;

    /** border.heightMargin(). */

    public int paneBorderHeight = 0;

    /** Width of the left pane (determined by the TOPLEFT pane). */

    public int leftMargin = 0;

    /** Width of the right pane (determined by the BOTTOMRIGHT pane). */

    public int rightMargin = 0;

    /** Height of the top pane (determined by the TOPLEFT pane). */

    public int topMargin = 0;

    /** Height of the bottom pane (determined by teh BOTTOMRIGHT pane). */

    public int bottomMargin = 0;

    /** ContainerView boundaries in X direction. */

    public final int viewX[] = new int[4];

    /** ContainerView boundaries in Y direction. */

    public final int viewY[] = new int[4];

    /** ContainerView width for each pane. */

    public final int viewWidth[] = new int[3];

    /** ContainerView height for each pane. */

    public final int viewHeight[] = new int[3];

    /** Temporary subgrid. */
    
    final Rect tempSubGrid = new Rect();

    /** Temporary rectangle. */
    
    final Rect tempRect = new Rect();

    /** Set to true if horizontal scroll bar is visible. */
    
    public boolean hasHorizScrollBar = false;

    /** Set to true if vertical scroll bar is visible. */

    public boolean hasVertScrollBar = false;

    /** Scroll Bar size. */

    public int scrollBarSize = 18;

    /** Creates IGridPane for the given pane index.  Override this to use customized IGridPanes. */

    public IGridPane createIGridPane(int paneIndex)
    {
        return new IGridPane(iGridDef, paneIndex);
    }

    /** Creates ContainerView for the given pane index.  Override this to use customized ContainerViews. */

    public ContainerView createContainerView(int paneIndex)
    {
        return new ContainerView();
    }

    /** Creates ScrollView for the given pane index.  Override this to use customized ScrollViews. */

    public ScrollView createScrollView(int paneIndex)
    {
        return new ScrollView();
    }

    /** Sets IGridDef object.  All the subviews (panes, scroll bars) are
        created here. */

    public void setIGridDef(IGridDef iGridDef)
    {
        this.iGridDef = iGridDef;

        for (int pane = 0; pane < 9; pane ++)
        {
            IGridPane iGridPane = createIGridPane(pane);
            ScrollView scrollView = createScrollView(pane);
            ContainerView containerView = createContainerView(pane);

            containerView.setBorder(paneBorder);

            iGridPane.iGridView = this;

            scrollView.setContentView(iGridPane);
            scrollView.setTransparent(true);

            scrollViewArray[pane] = scrollView;
            iGridPaneArray[pane] = iGridPane;
            containerViewArray[pane] = containerView;
            
            containerView.addSubview(scrollView);
            addSubview(containerView);
        }

        computePaneBorderMargin();

        ScrollView main = scrollViewArray[MIDDLECENTER];

        horizScrollBar = new ScrollBar(0, 0, 0, 0, Scrollable.HORIZONTAL);
        vertScrollBar = new ScrollBar(0, 0, 0, 0, Scrollable.VERTICAL);

        horizScrollBar.setScrollableObject(main);
        vertScrollBar.setScrollableObject(main);

        main.addScrollBar(horizScrollBar);
        main.addScrollBar(vertScrollBar);

        main.addScrollBar(this);

        addSubview(horizScrollBar);
        addSubview(vertScrollBar);
    }

    /** Creates an IGridView with a given IGridDef. */

    public IGridView(IGridDef iGridDef)
    {
        super();

        if (iGridDef != null)
        {
            setIGridDef(iGridDef);
        }

        this.setLayoutManager(this);
    }

    /** Creates an IGridView without any IGridDef. */

    public IGridView()
    {
        this(null);
    }
    
    /** Sets pane border. */

    public void setPaneBorder(Border paneBorder)
    {
        this.paneBorder = paneBorder;

        for (int pane = 0; pane < 9; pane ++)
        {
            containerViewArray[pane].setBorder(paneBorder);
        }

        computePaneBorderMargin();

        layoutView(0, 0);
    }

    /** Computes pane border margins. */

    public void computePaneBorderMargin()
    {
        paneBorderWidth = 0;
        paneBorderHeight = 0;

        if (paneBorder != null)
        {
            paneBorderWidth = paneBorder.widthMargin();
            paneBorderHeight = paneBorder.heightMargin();
        }
    }

    /** Computes layout. */

    public void computeLayout()
    {
        Rect topLeft = iGridDef.getPaneSubGrid(TOPLEFT);

        leftMargin = iGridDef.getXPosition(topLeft.x + topLeft.width) - iGridDef.getXPosition(topLeft.x);
        topMargin = iGridDef.getYPosition(topLeft.y + topLeft.height) - iGridDef.getYPosition(topLeft.y);

        Rect bottomRight = iGridDef.getPaneSubGrid(BOTTOMRIGHT);

        rightMargin = iGridDef.getXPosition(bottomRight.x + bottomRight.width) - iGridDef.getXPosition(bottomRight.x);
        bottomMargin = iGridDef.getYPosition(bottomRight.y + bottomRight.height) - iGridDef.getYPosition(bottomRight.y);

        if (leftMargin > 0)
        {
            leftMargin += paneBorderWidth;
        }

        if (rightMargin > 0)
        {
            rightMargin += paneBorderWidth;
        }

        if (topMargin > 0)
        {
            topMargin += paneBorderHeight;
        }
        
        if (bottomMargin > 0)
        {
            bottomMargin += paneBorderHeight;
        }

        int mainWidth = width() - leftMargin - rightMargin;
        int mainHeight = height() - topMargin - bottomMargin;

        int mainInteriorWidth = mainWidth - paneBorderWidth;
        int mainInteriorHeight = mainHeight - paneBorderHeight;

        Rect middleCenter = iGridDef.getPaneSubGrid(MIDDLECENTER);

        int cWidth = iGridDef.getXPosition(middleCenter.x + middleCenter.width) - iGridDef.getXPosition(middleCenter.x);
        int cHeight = iGridDef.getYPosition(middleCenter.y + middleCenter.height) - iGridDef.getYPosition(middleCenter.y);

        if (cWidth > mainInteriorWidth)
        {
            hasHorizScrollBar = true;
        }
        else
        {
            hasHorizScrollBar = false;
        }

        if (cHeight > mainInteriorHeight)
        {
            hasVertScrollBar = true;
        }
        else
        {
            hasVertScrollBar = false;
        }

        if (hasHorizScrollBar)
        {
            if (cHeight > mainInteriorHeight - scrollBarSize)
            {
                hasVertScrollBar = true;
            }
        }

        if (hasVertScrollBar)
        {
            if (cWidth > mainInteriorWidth - scrollBarSize)
            {
                hasHorizScrollBar = true;
            }
        }

        viewX[0] = 0;
        viewX[1] = leftMargin;
        viewX[3] = width() - ((hasVertScrollBar) ? scrollBarSize : 0);
        viewX[2] = viewX[3] - rightMargin;

        viewY[0] = 0;
        viewY[1] = topMargin;
        viewY[3] = height() - ((hasHorizScrollBar) ? scrollBarSize : 0);
        viewY[2] = viewY[3] - bottomMargin;

        for (int i = 0; i < 3; i ++)
        {
            viewWidth[i] = viewX[i + 1] - viewX[i];
            viewHeight[i] = viewY[i + 1] - viewY[i];
        }
    }

    // LayoutManager

    public void addSubview(View aView)
    {
        super.addSubview(aView);
    }

    public void removeSubview(View aView)
    {
        super.removeSubview(aView);
    }

    public void layoutView(View aView, int deltaWidth, int deltaHeight)
    {
        if (iGridDef == null)
        {
            return;
        }

        computeLayout();

        boolean isContainerDrawn = false;

        for (int i = 0; i < 3; i ++)
        {
            int vX = viewX[i];
            int vWidth = viewWidth[i];

            for (int j = 0; j < 3; j ++)
            {
                int vY = viewY[j];
                int vHeight = viewHeight[j];

                int pane = i + j * 3;

                Rect rectNow = containerViewArray[pane].bounds();

                tempRect.setBounds(vX, vY, vWidth, vHeight);

                if (!rectNow.equals(tempRect))
                {
                    containerViewArray[pane].setBounds(tempRect);

                    Rect r = containerViewArray[pane].interiorRect();

                    scrollViewArray[pane].setBounds(r);

                    containerViewArray[pane].setDirty(true);

                    isContainerDrawn = true;
                }
            }
        }

        if (isContainerDrawn)
        {
            if (hasHorizScrollBar)
            {
                horizScrollBar.setBounds(viewX[1], viewY[3], viewWidth[1], scrollBarSize);
                tempRect.setBounds(0, horizScrollBar.y(), width(), horizScrollBar.height());
                addDirtyRect(tempRect);
            }
            else
            {
                horizScrollBar.sizeTo(0, 0);
            }
            
            if (hasVertScrollBar)
            {
                vertScrollBar.setBounds(viewX[3], viewY[1], scrollBarSize, viewHeight[1]);
                tempRect.setBounds(vertScrollBar.x(), 0, vertScrollBar.height(), height());
                addDirtyRect(tempRect);
            }
            else
            {
                vertScrollBar.sizeTo(0, 0);
            }
        }
    }

    /** Updates cells in subGrid (in all panes) after processing current event. */

    public void updateCells(Rect subGrid)
    {
        for (int pane = 0; pane < 9; pane ++)
        {
            iGridPaneArray[pane].updateCells(subGrid);
        }

        layoutView(0, 0);
    }

    /** Updates cells in subGrid (in all panes) immediately. */

    public void updateCellsNow(Rect subGrid)
    {
        for (int pane = 0; pane < 9; pane ++)
        {
            iGridPaneArray[pane].updateCellsNow(subGrid);
        }

        layoutView(0, 0);
    }

    /** Extends a subGrid to contain all the extended subcells. */

    public void extendSubGrid(Rect subGrid)
    {
        boolean isExtended = false;

        do
        {
            isExtended = false;

        recompute:

            for (int y = subGrid.y; y < subGrid.y + subGrid.height; y ++)
            {
                for (int x = subGrid.x; x < subGrid.x + subGrid.width; x ++)
                {
                    tempSubGrid.setBounds(x, y, 1, 1);

                    if (iGridDef.extendCell(tempSubGrid))
                    {
                        if (!subGrid.contains(tempSubGrid))
                        {
                            subGrid.unionWith(tempSubGrid);
                            
                            isExtended = true;
                            break recompute;
                        }
                    }
                }
            }
        }
        while (isExtended);
    }

    /** Allocates the panes as follows (L=leftMargin, R=rightMargin, T=topMargin, B=bottomMargin, X=XCount, Y=YCount):<BR>
        <P>
        <UL>
        <LI>TOPLEFT: (0, 0, L, T)
        <LI>TOPCENTER: (L + R, 0, X - (L + R), T)
        <LI>TOPRIGHT: (L, 0, R, T)
        <LI>MIDDLELEFT: (0, T + B, L, Y - (T + B))
        <LI>MIDDLECENTER: (L + R, T + B, X - (L + R), Y - (T + B))
        <LI>MIDDLERIGHT: (L, T + B, R, Y - (T + B))
        <LI>BOTTOMLEFT: (0, T, L, B)
        <LI>BOTTOMCENTER: (L + R, T, X - (L + R), B)
        <LI>BOTTOMRIGHT: (L, T, R, B)

        </UL>
    */

    public static Rect computeStandardPaneSubGrid(IGridDef iGridDef, int paneIndex, int leftMargin, int rightMargin, int topMargin, int bottomMargin, Rect paneRect)
    {
        int yLevel = paneIndex / 3;
        int xLevel = paneIndex % 3;

        switch (xLevel)
        {
            case 0:
            {
                paneRect.x = 0;
                paneRect.width = leftMargin;
            }
            break;

            case 1:
            {
                paneRect.x = leftMargin + rightMargin;
                paneRect.width = iGridDef.getXCount() - (leftMargin + rightMargin);
            }
            break;

            case 2:
            {
                paneRect.x = leftMargin;
                paneRect.width = rightMargin;
            }
            break;
        }

        switch (yLevel)
        {
            case 0:
            {
                paneRect.y = 0;
                paneRect.height = topMargin;
            }
            break;

            case 1:
            {
                paneRect.y = topMargin + bottomMargin;
                paneRect.height = iGridDef.getYCount() - (topMargin + bottomMargin);
            }
            break;

            case 2:
            {
                paneRect.y = topMargin;
                paneRect.height = bottomMargin;
            }
            break;
        }

        return paneRect;
    }

    /** Scrolls the main view to show the cell at the index. */

    public void scrollCellToVisible(Point index)
    {
        IGridPane mainContent = iGridPaneArray[MIDDLECENTER];
        ScrollView main = scrollViewArray[MIDDLECENTER];

        int x = index.x;
        int y = index.y;

        Rect mainIndexRect = iGridDef.getPaneSubGrid(MIDDLECENTER);

        if (x < mainIndexRect.x)
        {
            x = mainIndexRect.x;
        }
        else if (x > mainIndexRect.maxX())
        {
            x = mainIndexRect.maxX();
        }

        if (y < mainIndexRect.y)
        {
            y = mainIndexRect.y;
        }
        else if (y > mainIndexRect.maxY())
        {
            y = mainIndexRect.maxY();
        }

        tempSubGrid.setBounds(x, y, 1, 1);

        mainContent.computeCellRect(tempSubGrid, tempRect);

        Rect cellRect = mainContent.convertRectToView(main, tempRect);

        main.scrollRectToVisible(cellRect);
    }

    /** Resizes this window so that the whole grid is displayed (without scrolling). */

    public void sizeToIGridSize()
    {
        computeLayout();

        Rect middleCenter = iGridPaneArray[MIDDLECENTER].bounds();

        int cWidth = middleCenter.width;
        int cHeight = middleCenter.height;

        sizeTo(cWidth + paneBorderWidth + leftMargin + rightMargin, cHeight + paneBorderHeight + topMargin + bottomMargin);
    }

    /** Handles multi-pane scrolling. */

    public void performCommand(String command, Object object)
    {
        if (command == null)
        {
            return;
        }

        if (command.equals(ScrollBar.UPDATE))
        {
            ScrollView source = (ScrollView)object;

            int x = source.contentView().x();
            int y = source.contentView().y();

            scrollViewArray[TOPCENTER].scrollTo(x, 0);
            scrollViewArray[BOTTOMCENTER].scrollTo(x, 0);

            scrollViewArray[MIDDLELEFT].scrollTo(0, y);
            scrollViewArray[MIDDLERIGHT].scrollTo(0, y);
        }
    }
}




