// Delegated View

// Copyright (C) 1999, 2000 by Makoto Nagata, MicroNova <mnagata@micronova.com>.  
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.
//
// Visit http://www.micronova.com for updates

package com.micronova.mnfc;

import netscape.application.*;

/**

   IView is a View that dynamically delegates selected method calls (such as mouseDown(), drawView(), etc.) to its IViewDelegate object.  Each method has its own bit flag (e.g., MOUSE_DOWN for mouseDown() method) to enable delegation.  If the bit flag for a method is set (by setDelegateFlag()), whenever the method is called,  the delegate object's delegate() method is called with the bit flag and the IView itself.  Method parameters and return value are both instance variables of IView (parameters have '__' prefix, return values have '_' prefix) so IViewDelegate can examine method parameters and/or modify the return value during delegation.

   @author Makoto Nagata, MicroNova
   @version 1.0

*/

public class IView extends View implements IViewDelegate, IToolTipSource
{
    /** IViewDelegate object.  Defaults to IView itself. */

    public IViewDelegate iViewDelegate;

    /** Defines which methods to be delegated (sum of bit flags). */

    public int delegateFlag = 0;

    /** Return value of mouseDown method. */
    
    public boolean _mouseDown = false;

    /** Return value of wantsAutoScrollEvents(). */

    public boolean _wantsAutoscrollEvents = false;

    /** Return value of wantsMouseEventCoalescing(). */

    public boolean _wantsMouseEventCoalescing = true;

    /** Return value of isTransparent(). */

    public boolean _isTransparent = false;

    /** Return value of acceptsDrag(). */

    public DragDestination _acceptsDrag = null;

    /** Return value of cursorForPoint(). */

    public int _cursorForPoint = View.ARROW_CURSOR;

    /** Return value of nextSelectableView(). */

    public View _nextSelectableView = null;

    /** Return value of previousSelectableView(). */

    public View _previousSelectableView = null;

    /** Return value of canBecomeSelectedView(). */

    public boolean _canBecomeSelectedView = false;

    /** Return value of hidesSubviewsFromKeyboard(). */

    public boolean _hidesSubviewsFromKeyboard = false;

    /** Return value of getToolTipText(). */

    public String _getToolTipText = null;

    /** Method parameter for MOUSE_XXX and KEY_XXX. */

    public Event __event;

    /** Method parameter for DRAW_VIEW and SUBVIEWS_DRAWN. */

    public Graphics __g;

    /** Method parameter for SUBVIEW_XXX. */

    public View __view;

    /** Method parameter for DID_MOVE_BY, DID_SIZE_BY, CURSOR_FOR_POINT, ACCEPTS_DRAG. */

    public int __x;

    /** Method parameter for DID_MOVE_BY, DID_SIZE_BY, CURSOR_FOR_POINT, ACCEPTS_DRAG. */

    public int __y;

    /** Method parameter for ACCEPTS_DRAG. */

    public DragSession __session;

    /** Free variable */

    public Object __object;

    /** flag to delegate mouseDown(). */

    public final static int MOUSE_DOWN    = 0x00000001;

    /** flag to delegate mouseUp(). */

    public final static int MOUSE_UP      = 0x00000002;

    /** flag to delegate mouseMoved(). */

    public final static int MOUSE_MOVED   = 0x00000004;

    /** flag to delegate mouseEntered(). */

    public final static int MOUSE_ENTERED = 0x00000008;

    /** flag to delegate mouseExited(). */

    public final static int MOUSE_EXITED  = 0x00000010;

    /** flag to delegate mouseDragged(). */

    public final static int MOUSE_DRAGGED = 0x00000020;

    /** flag to delegate all mouseXXX() methods. */

    public final static int MOUSE         = 0x0000003F;

    /** flag to delegate keyDown(). */

    public final static int KEY_DOWN      = 0x00000040;

    /** flag to delegate keyUp(). */

    public final static int KEY_UP        = 0x00000080;

    /** flag to delegate all KeyXXX() methods. */

    public final static int KEY           = 0x000000C0;

    /** flag to delegate startFocus(). */

    public final static int START_FOCUS   = 0x00000100;

    /** flag to delegate stopFocus(). */

    public final static int STOP_FOCUS    = 0x00000200;

    /** flag to delegate pauseFocus(). */

    public final static int PAUSE_FOCUS   = 0x00000400;

    /** flag to delegate resumeFocus(). */

    public final static int RESUME_FOCUS  = 0x00000800;

    /** flag to delegate all focusXXX(). */

    public final static int FOCUS         = 0x00000F00;

    /** flag to delegate notification after super.ancestorWasAddedToViewHierarchy(). */

    public final static int ANCESTOR_WAS_ADDED_TO_VIEW_HIERARCHY = 0x00001000;

    /** flag to delegate notification before super.ancestorWillRemoveFromViewHierarchy(). */

    public final static int ANCESTOR_WILL_REMOVE_FROM_VIEW_HIERARCHY = 0x000020000;

    /** flag to delegate all ANCESTOR_xxx notifications. */

    public final static int ANCESTOR = 0x00003000;

    /** flag to delegate subviewDidResize(). */

    public final static int SUBVIEW_DID_RESIZE     = 0x00004000;

    /** flag to delegate subviewDidMove(). */

    public final static int SUBVIEW_DID_MOVE       = 0x00008000;

    /** flag to delegate all subviewXXX(). */

    public final static int SUBVIEW                = 0x0000C000;

    /** flag to delegate didMoveBy(). */

    public final static int DID_MOVE_BY = 0x00010000;

    /** flag to delegate didSizeBy(). */

    public final static int DID_SIZE_BY = 0x00020000;

    /** flag to delegate drawView(). */

    public final static int DRAW_VIEW = 0x00040000;

    /** flag to delegate notification after drawSubviews(). */

    public final static int SUBVIEWS_DRAWN = 0x00080000;

    /** flag to delegate acceptsDrag(). */

    public final static int ACCEPTS_DRAG = 0x00100000;

    /** flag to delegate cursorForPoint(). */

    public final static int CURSOR_FOR_POINT = 0x00200000;

    /** flag to delegate willBecomeSelected(). */

    public final static int WILL_BECOME_SELECTED   = 0x00400000;

    /** flag to delegate willBecomeUnselected(). */

    public final static int WILL_BECOME_UNSELECTED = 0x00800000;

    /** flag to delegate nextSelectableView(). */

    public final static int NEXT_SELECTABLE_VIEW = 0x01000000;

    /** flag to delegate previousSelectableView(). */

    public final static int PREVIOUS_SELECTABLE_VIEW = 0x02000000;

    /** flag to delegate getToolTipText(). */

    public final static int GET_TOOL_TIP_TEXT = 0x04000000;

    /** flag to delegate wantsAutoscrollEvents() */

    public final static int WANTS_AUTOSCROLL_EVENTS = 0x08000000;

    /** flag to delegate wantsMouseEventColalescing(). */

    public final static int WANTS_MOUSE_EVENT_COALESCING = 0x10000000;

    /** flag to delegate wantsMouseEventColalescing(). */

    public final static int IS_TRANSPARENT = 0x20000000;

    /** flag to delegate canBecomeSelectedView(). */

    public final static int CAN_BECOME_SELECTED_VIEW = 0x40000000;

    /** flag to delegate hidesSubviewsFromKeyboard(). */

    public final static int HIDES_SUBVIEWS_FROM_KEYBOARD = 0x80000000;

    /** Sets IViewDelegate object. */
    
    public void setIViewDelegate(IViewDelegate delegate)
    {
        this.iViewDelegate = delegate;
    }

    /** Sets the delegateFlag (sum of bit flags); e.g., setDelegateFlag(MOUSE + KEY) to delegate all mouseXXX and keyXXX methods. */

    public void setDelegateFlag(int delegateFlag)
    {
        this.delegateFlag = delegateFlag;
    }

    /** Sets/clears individual delegate flag (e.g., setDelegateFlag(MOUSE_DOWN, true) to delegate mouseDown(). */

    public void setDelegateFlag(int delegateFlag, boolean enable)
    {
        if (enable)
        {
            this.delegateFlag |= delegateFlag;
        }
        else
        {
            this.delegateFlag &= (~delegateFlag);
        }
    }

    /** Constructs an IView with bounds (x, y, width, height) and sets itself as the delegate object. */

    public IView(int x, int y, int width, int height)
    {
        super(x, y, width, height);

        setIViewDelegate(this);
    }

    /** Constructs an IView with bounds r and sets itself as the delegate object.  */

    public IView(Rect r)
    {
        this(r.x, r.y, r.width, r.height);
    }

    /** Constructs an IView with bounds (0, 0, 0, 0) and sets itself as the delegate object.  */

    public IView()
    {
        this(0, 0, 0, 0);
    }

    /** If MOUSE_DOWN flag is set, sets '__event' to the MouseEvent and delegates.   Returns _mouseDown variable. */

    public boolean mouseDown(MouseEvent event)
    {
        if ((delegateFlag & MOUSE_DOWN) != 0)
        {
            this.__event = event;
            iViewDelegate.delegate(MOUSE_DOWN, this);
        }

        return _mouseDown;
    }

    /** If MOUSE_MOVED flag is set, sets '__event' to the MouseEvent and delegates. */

    public void mouseMoved(MouseEvent event)
    {
        if ((delegateFlag & MOUSE_MOVED) != 0)
        {
            this.__event = event;
            iViewDelegate.delegate(MOUSE_MOVED, this);
        }
    }

    /** If MOUSE_UP flag is set, sets '__event' to the MouseEvent and delegates. */

    public void mouseUp(MouseEvent event)
    {
        if ((delegateFlag & MOUSE_UP) != 0)
        {
            this.__event = event;
            iViewDelegate.delegate(MOUSE_UP, this);
        }
    }

    /** If MOUSE_DRAGGED flag is set, sets '__event' to the MouseEvent and delegates. */

    public void mouseDragged(MouseEvent event)
    {
        if ((delegateFlag & MOUSE_DRAGGED) != 0)
        {
            this.__event = event;
            iViewDelegate.delegate(MOUSE_DRAGGED, this);
        }
    }

    /** If MOUSE_ENTERED flag is set, sets '__event' to the MouseEvent and delegates. */

    public void mouseEntered(MouseEvent event)
    {
        if ((delegateFlag & MOUSE_ENTERED) != 0)
        {
            this.__event = event;
            iViewDelegate.delegate(MOUSE_ENTERED, this);
        }
    }

    /** If MOUSE_EXITED flag is set, sets '__event' to the MouseEvent and delegates. */

    public void mouseExited(MouseEvent event)
    {
        if ((delegateFlag & MOUSE_EXITED) != 0)
        {
            this.__event = event;
            iViewDelegate.delegate(MOUSE_EXITED, this);
        }
    }

    /** If KEY_DOWN flag is set, sets '__event' to the KeyEvent and delegates.  Note that there is no default implementation (so the event is not automatically passed to the superview).  */

    public void keyDown(KeyEvent event)
    {
        if ((delegateFlag & KEY_DOWN) != 0)
        {
            this.__event = event;

            iViewDelegate.delegate(KEY_DOWN, this);
        }
    }

    /** If KEY_UP flag is set, sets '__event' to the KeyEvent and delegates.   Note that there is no default implementation (so the event is not automatically passed to the superview). */

    public void keyUp(KeyEvent event)
    {
        if ((delegateFlag & KEY_UP) != 0)
        {
            this.__event = event;
            iViewDelegate.delegate(KEY_UP, this);
        }
    }

    /** If START_FOCUS flag is set, delegates. */

    public void startFocus()
    {
        if ((delegateFlag & START_FOCUS) != 0)
        {
            iViewDelegate.delegate(START_FOCUS, this);
        }
    }

    /** If STOP_FOCUS flag is set, delegates. */

    public void stopFocus()
    {
        if ((delegateFlag & STOP_FOCUS) != 0)
        {
            iViewDelegate.delegate(STOP_FOCUS, this);
        }
    }

    /** If PAUSE_FOCUS flag is set, delegates. */

    public void pauseFocus()
    {
        if ((delegateFlag & PAUSE_FOCUS) != 0)
        {
            iViewDelegate.delegate(PAUSE_FOCUS, this);
        }
    }

    /** If RESUME_FOCUS flag is set, delegates. */

    public void resumeFocus()
    {
        if ((delegateFlag & RESUME_FOCUS) != 0)
        {
            iViewDelegate.delegate(RESUME_FOCUS, this);
        }
    }

    /** If ANCESTOR_WAS_ADDED_TO_VIEW_HIERARCHY flag is set, delegates after calling the View's default implementation. */

    public void ancestorWasAddedToViewHierarchy(View addedView)
    {
        super.ancestorWasAddedToViewHierarchy(addedView);

        if ((delegateFlag & ANCESTOR_WAS_ADDED_TO_VIEW_HIERARCHY) != 0)
        {
            this.__view = addedView;
            iViewDelegate.delegate(ANCESTOR_WAS_ADDED_TO_VIEW_HIERARCHY, this);
        }
    }

    /** If ANCESTOR_WILL_REMOVE_FROM_VIEW_HIERARCHY flag is set, delegates before calling the View's default implementation. */

    public void ancestorWillRemoveFromViewHierarchy(View removedView)
    {
        if ((delegateFlag & ANCESTOR_WILL_REMOVE_FROM_VIEW_HIERARCHY) != 0)
        {
            this.__view = removedView;
            iViewDelegate.delegate(ANCESTOR_WILL_REMOVE_FROM_VIEW_HIERARCHY, this);
        }

        super.ancestorWillRemoveFromViewHierarchy(removedView);
    }

    /** If DRAW_VIEW flag is set, sets '__g' to the Graphics and delegates. */

    public void drawView(Graphics g)
    {
        if ((delegateFlag & DRAW_VIEW) != 0)
        {
            this.__g = g;
            iViewDelegate.delegate(DRAW_VIEW, this);
        }
    }

    /** If SUBVIEWS_DRAWN flag is set, sets '__g' to the Graphics and delegates after calling the View's default implementation. */

    public void drawSubviews(Graphics g)
    {
        super.drawSubviews(g);

        if ((delegateFlag & SUBVIEWS_DRAWN) != 0)
        {
            this.__g = g;
            iViewDelegate.delegate(SUBVIEWS_DRAWN, this);
        }
    }

    /** If SUBVIEW_DID_RESIZE flag is set, sets '__view' to the subview and delegates. */

    public void subviewDidResize(View aSubview)
    {
        if ((delegateFlag & SUBVIEW_DID_RESIZE) != 0)
        {
            this.__view = aSubview;
            iViewDelegate.delegate(SUBVIEW_DID_RESIZE, this);
        }
    }

    /** If SUBVIEW_DID_MOVE flag is set, sets '__view' to the subview and delegates. */

    public void subviewDidMove(View aSubview)
    {
        if ((delegateFlag & SUBVIEW_DID_MOVE) != 0)
        {
            this.__view = aSubview;
            iViewDelegate.delegate(SUBVIEW_DID_MOVE, this);
        }
    }

    /** Returns the value of _wantsMouseEventCoalescing. */

    public boolean wantsMouseEventCoalescing()
    {
        if ((delegateFlag & WANTS_MOUSE_EVENT_COALESCING) != 0)
        {
            iViewDelegate.delegate(WANTS_MOUSE_EVENT_COALESCING, this);
        }

        return _wantsMouseEventCoalescing;
    }

    /** Returns the value of _wantsAutoscrollEvents. */

    public boolean wantsAutoscrollEvents()
    {
        if ((delegateFlag & WANTS_AUTOSCROLL_EVENTS) != 0)
        {
            iViewDelegate.delegate(WANTS_AUTOSCROLL_EVENTS, this);
        }

        return _wantsAutoscrollEvents;
    }

    /** If DID_MOVE_BY flag is set, sets '__x' to deltaX and '__y' to deltaY and delegates. */

    public void didMoveBy(int deltaX, int deltaY)
    {
        if ((delegateFlag & DID_MOVE_BY) != 0)
        {
            this.__x = deltaX;
            this.__y = deltaY;
            iViewDelegate.delegate(DID_MOVE_BY, this);
        }
    }

    /** If DID_SIZE_BY flag is set, sets '__x' to deltaWidth and '__y' to deltaHeight and delegates. */

    public void didSizeBy(int deltaWidth, int deltaHeight)
    {
        if ((delegateFlag & DID_SIZE_BY) != 0)
        {
            this.__x = deltaWidth;
            this.__y = deltaHeight;
            iViewDelegate.delegate(DID_SIZE_BY, this);
        }
    }

    /** If ACCEPTS_DRAG flag is set, sets '__session' to session and '__x' to x and '__y' to y and delegates.  Returns _acceptsDrag. */

    public DragDestination acceptsDrag(DragSession session, int x, int y)
    {
        if ((delegateFlag & ACCEPTS_DRAG) != 0)
        {
            this.__session = session;
            this.__x = x;
            this.__y = y;

            iViewDelegate.delegate(ACCEPTS_DRAG, this);
        }

        return _acceptsDrag;
    }

    /** If CURSOR_FOR_POINT flag is set, sets and '__x' to x and '__y' to y and delegates.   Always returns the value of _cursorForPoint. */

    public int cursorForPoint(int x, int y)
    {
        if ((delegateFlag & CURSOR_FOR_POINT) != 0)
        {
            this.__x = x;
            this.__y = y;
            iViewDelegate.delegate(CURSOR_FOR_POINT, this);
        }

        return _cursorForPoint;
    }

    /** If WILL_BECOME_SELECTED flag is set, delegates. */

    public void willBecomeSelected()
    {
        if ((delegateFlag & WILL_BECOME_SELECTED) != 0)
        {
            iViewDelegate.delegate(WILL_BECOME_SELECTED, this);
        }
    }

    /** If WILL_BECOME_UNSELECTED flag is set, delegates. */

    public void willBecomeUnselected()
    {
        if ((delegateFlag & WILL_BECOME_UNSELECTED) != 0)
        {
            iViewDelegate.delegate(WILL_BECOME_UNSELECTED, this);
        }
    }

    /** Returns the value of _isTransparent. */

    public boolean isTransparent()
    {
        if ((delegateFlag & IS_TRANSPARENT) != 0)
        {
            iViewDelegate.delegate(IS_TRANSPARENT, this);
        }

        return _isTransparent;
    }

    /** Returns the value of _canBecomeSelectedView. */

    public boolean canBecomeSelectedView()
    {
        if ((delegateFlag & CAN_BECOME_SELECTED_VIEW) != 0)
        {
            iViewDelegate.delegate(CAN_BECOME_SELECTED_VIEW, this);
        }

        return _canBecomeSelectedView;
    }

    /** Returns the value of _hidesSubviewsFromKeyboard. */

    public boolean hidesSubviewsFromKeyboard()
    {
        if ((delegateFlag & HIDES_SUBVIEWS_FROM_KEYBOARD) != 0)
        {
            iViewDelegate.delegate(HIDES_SUBVIEWS_FROM_KEYBOARD, this);
        }

        return _hidesSubviewsFromKeyboard;
    }

    /** If NEXT_SELECTABLE_VIEW flag is set, delegates.  Always returns the value of _nextSelectedView. */

    public View nextSelectableView()
    {
        if ((delegateFlag & NEXT_SELECTABLE_VIEW) != 0)
        {
            iViewDelegate.delegate(NEXT_SELECTABLE_VIEW, this);
        }

        return this._nextSelectableView;
    }

    /** If PREVIOUS_SELECTABLE_VIEW flag is set, delegates.  Always returns the value of _previousSelectedView. */

    public View previousSelectableView()
    {
        if ((delegateFlag & PREVIOUS_SELECTABLE_VIEW) != 0)
        {
            iViewDelegate.delegate(PREVIOUS_SELECTABLE_VIEW, this);
        }

        return this._previousSelectableView;
    }

    /** If GET_TOOL_TIP_TEXT flag is set, delegates.  Returns the value of _getToolTipText. */

    public String getToolTipText(MouseEvent event)
    {
        if ((delegateFlag & GET_TOOL_TIP_TEXT) != 0)
        {
            __event = event;
            iViewDelegate.delegate(GET_TOOL_TIP_TEXT, this);
        }

        return this._getToolTipText;
    }


    /** Dummy self-delegation.  This does nothing. */

    public void delegate(int flag, IView iView)
    {
    }
}
